module Database where

import Core

import Control.Exception            (IOException)
import Data.Aeson                   (ToJSON)
import Data.ByteString              (ByteString)
import Data.Maybe                   (listToMaybe)
import Data.Pool
import qualified Data.Text          as T
import Database.PostgreSQL.Simple
import Database.PostgreSQL.Simple.ToField (ToField)
import Effectful
import Effectful.Dispatch.Dynamic
import Effectful.Error.Static       (Error)
import Effectful.Reader.Static
import Effectful.State.Static.Local (State, get, put, evalState)
import Servant hiding               ((:>), throwError)

type instance DispatchOf Database = 'Dynamic

type DatabaseExeEffects es = ( Reader AppEnv :> es 
                             , Error ServerError :> es
                             , IOE :> es
                             )

databaseInit :: (Database :> es, Error ServerError :> es) 
             => Eff es ()
databaseInit = send DatabaseInit

databaseRead :: (ToField a, Show a, Database :> es, Error ServerError :> es)
             => (Query, a) -> Eff es [User]
databaseRead = send . DatabaseRead

databaseRead_ :: (Database :> es, Error ServerError :> es) 
              => Query -> Eff es [User]
databaseRead_ = send . DatabaseRead_

databaseWrite :: (ToRow a, Show a, Database :> es, Error ServerError :> es)
              => (Query, a) -> Eff es ()
databaseWrite = send . DatabaseWrite

runDatabaseIO :: DatabaseExeEffects es 
              => Eff (Database : es) a -> Eff es a
runDatabaseIO = interpret $ \_ -> \case
  DatabaseInit -> do
    AppEnv { pool } <- ask
    liftWithPool pool $ \conn -> do
      execute_ conn createUsersTable
      execute_ conn createLogsTable
      pure ()

  DatabaseRead (statement, values) -> do
    AppEnv { pool } <- ask
    liftWithPool pool $ \conn ->
      query conn statement (Only values)

  DatabaseRead_ statement -> do
    AppEnv { pool } <- ask
    liftWithPool pool $ \conn ->
      query_ conn statement

  DatabaseWrite (statement, values) -> do
    AppEnv { pool } <- ask
    liftWithPool pool $ \conn -> do
      execute conn statement values
      pure ()
  where
    liftWithPool :: (IOE :> es, Error ServerError :> es) 
                 => (Pool Connection) -> (Connection -> IO a) -> Eff es a
    liftWithPool p f = liftIOE $ withResource p f

runDatabaseDebug :: DatabaseExeEffects es 
                 => Eff (Database : es) a -> Eff es a
runDatabaseDebug = interpret $ \_ -> \case
  DatabaseInit -> do
    liftIOE $ putStrLn "Mocked setup of the database"
  DatabaseRead (statement, values) -> do
    liftIOE $ putStrLn
       $ "Mocked a READ database operation with statement:\n"
      ++ show statement
      ++ "\nValues:\n"
      ++ show values
    pure []
  DatabaseRead_ statement -> do
    liftIOE $ putStrLn "Mocked a READ database operation on all users"
    pure []
  DatabaseWrite (statement, values) -> do
    liftIOE $ putStrLn
      $ "Mocked a WRITE database operation with statement:\n"
     ++ show statement 
     ++ "\nValues:\n"
     ++ show values

createConnectionPool :: ByteString -> IO (Pool Connection)
createConnectionPool connectString = newPool $ defaultPoolConfig
  (connectPostgreSQL connectString)
  close
  60
  10

queryUser :: UserId -> (Query, UserId)
queryUser userId = ("SELECT id, name FROM users WHERE id = ?;", userId)

writeUser :: T.Text -> (Query, Only T.Text)
writeUser name = ("INSERT INTO users (name) VALUES (?);", Only name)

deleteUser :: UserId -> (Query, UserId)
deleteUser userId = ("DELETE FROM users WHERE id = ?;", userId)

createUsersTable :: Query
createUsersTable = "CREATE TABLE IF NOT EXISTS users ( \
                   \  id SERIAL PRIMARY KEY, \
                   \  name character varying(255) NOT NULL \
                   \);"

createLogsTable :: Query
createLogsTable = "CREATE TABLE IF NOT EXISTS logs ( \
                  \  id SERIAL PRIMARY KEY, \
                  \  level VARCHAR(10) NOT NULL, \
                  \  message TEXT NOT NULL, \
                  \  timestamp TIMESTAMPTZ NOT NULL DEFAULT NOW(), \
                  \  source VARCHAR(100), \
                  \  context JSONB \
                  \);"
